#!/usr/bin/env python

import copy
import json
from pathlib import Path

import boto3
import click
import yaml
from jsonschema import validate as validate_json

from .utils import SSM_BASE_PATH
from .utils import camel_case
from .utils import ensure_cwd_is_repo_root
from .utils import mkdir
from .utils import mkfile
from .utils import setup_templates

PACKAGE_DIR = Path(__file__).resolve().parent

WAF_ACL_ARN_KEY = "waf-acl-arn"


def list_copilot_local_environments():
    return [
        path.parent.parts[-1] for path in Path("./copilot/environments/").glob("*/manifest.yml")
    ]


def list_copilot_local_services():
    return [path.parent.parts[-1] for path in Path("./copilot/").glob("*/manifest.yml")]


@click.group()
def copilot():
    pass


def _validate_and_normalise_config(config_file):
    """Load the addons.yaml file, validate it and return the normalised config
    dict."""

    def _lookup_plan(addon_type, env_conf):
        plan = env_conf.pop("plan", None)
        conf = addon_plans[addon_type][plan] if plan else {}
        conf.update(env_conf)

        return conf

    def _normalise_keys(source: dict):
        return {k.replace("-", "_"): v for k, v in source.items()}

    with open(PACKAGE_DIR / "addon-plans.yml", "r") as fd:
        addon_plans = yaml.safe_load(fd)

    with open(PACKAGE_DIR / "schemas/addons-schema.json", "r") as fd:
        schema = json.load(fd)

    # load and validate config
    with open(config_file, "r") as fd:
        config = yaml.safe_load(fd)

    # empty file
    if not config:
        return {}

    validate_json(instance=config, schema=schema)

    env_names = list_copilot_local_environments()
    svc_names = list_copilot_local_services()

    if not env_names:
        click.echo(
            click.style(f"No environments found in ./copilot/environments; exiting", fg="red")
        )
        exit(1)

    if not svc_names:
        click.echo(click.style(f"No services found in ./copilot/; exiting", fg="red"))
        exit(1)

    normalised_config = {}
    for addon_name, addon_config in config.items():
        addon_type = addon_config["type"]
        normalised_config[addon_name] = copy.deepcopy(addon_config)

        if "services" in normalised_config[addon_name]:
            if type(normalised_config[addon_name]["services"]) == str:
                if normalised_config[addon_name]["services"] == "__all__":
                    normalised_config[addon_name]["services"] = svc_names
                else:
                    click.echo(
                        click.style(
                            f"{addon_name}.services must be a list of service names or '__all__'",
                            fg="red",
                        ),
                    )
                    exit(1)

            if not set(normalised_config[addon_name]["services"]).issubset(set(svc_names)):
                click.echo(
                    click.style(
                        f"Services listed in {addon_name}.services do not exist in ./copilot/",
                        fg="red",
                    ),
                )
                exit(1)

        environments = normalised_config[addon_name].pop("environments", {})
        default = environments.pop("default", {})

        initial = _lookup_plan(addon_type, default)

        if not set(environments.keys()).issubset(set(env_names)):
            click.echo(
                click.style(
                    f"Environment keys listed in {addon_name} do not match ./copilot/environments",
                    fg="red",
                ),
            )
            exit(1)

        normalised_environments = {}

        for env in env_names:
            normalised_environments[env] = _normalise_keys(initial)

        for env_name, env_config in environments.items():
            normalised_environments[env_name].update(
                _lookup_plan(addon_type, _normalise_keys(env_config))
            )

        normalised_config[addon_name]["environments"] = normalised_environments

    return normalised_config


@copilot.command()
def make_addons():
    """Generate addons CloudFormation for each environment."""

    overwrite = True
    output_dir = Path(".").absolute()

    ensure_cwd_is_repo_root()

    templates = setup_templates()

    config = _validate_and_normalise_config(PACKAGE_DIR / "default-addons.yml")
    project_config = _validate_and_normalise_config("addons.yml")
    config.update(project_config)

    with open(PACKAGE_DIR / "addons-template-map.yml") as fd:
        addon_template_map = yaml.safe_load(fd)

    click.echo("\n>>> Generating addons CloudFormation\n")

    path = Path(f"copilot/environments/addons/")
    mkdir(output_dir, path)

    services = []
    for addon_name, addon_config in config.items():
        print(f">>>>>>>>> {addon_name}")
        addon_type = addon_config.pop("type")
        environments = addon_config.pop("environments")

        environment_addon_config = {
            "secret_name": addon_name.upper().replace("-", "_"),
            "name": addon_config.get("name", None) or addon_name,
            "environments": environments,
            "prefix": camel_case(addon_name),
            "addon_type": addon_type,
            **addon_config,
        }

        services.append(environment_addon_config)

        service_addon_config = {
            "name": addon_config.get("name", None) or addon_name,
            "prefix": camel_case(addon_name),
            "environments": environments,
            **addon_config,
        }

        # generate env addons
        for addon in addon_template_map[addon_type].get("env", []):
            template = templates.get_template(addon["template"])

            contents = template.render({"service": environment_addon_config})

            filename = addon.get("filename", f"{addon_name}.yml")

            click.echo(mkfile(output_dir, path / filename, contents, overwrite=overwrite))

        # generate svc addons
        for addon in addon_template_map[addon_type].get("svc", []):
            template = templates.get_template(addon["template"])

            for svc in addon_config.get("services", []):
                service_path = Path(f"copilot/{svc}/addons/")

                contents = template.render({"service": service_addon_config})

                filename = addon.get("filename", f"{addon_name}.yml")

                mkdir(output_dir, service_path)
                click.echo(
                    mkfile(output_dir, service_path / filename, contents, overwrite=overwrite)
                )

        if addon_type in ["aurora-postgres", "rds-postgres"]:
            click.secho(
                "\nNote: The key DATABASE_CREDENTIALS may need to be changed to match your Django settings configuration.",
                fg="yellow",
            )

    click.echo(templates.get_template("addon-instructions.txt").render(services=services))


@copilot.command()
@click.argument("app", type=str)
@click.argument("env", type=str)
def get_env_secrets(app, env):
    """List secret names and values for an environment."""

    client = boto3.client("ssm")

    path = SSM_BASE_PATH.format(app=app, env=env)

    params = dict(Path=path, Recursive=False, WithDecryption=True, MaxResults=10)
    secrets = []

    # TODO: refactor shared code with get_ssm_secret_names
    while True:
        response = client.get_parameters_by_path(**params)

        for secret in response["Parameters"]:
            secrets.append(f"{secret['Name']:<8}: {secret['Value']:<15}")

        if "NextToken" in response:
            params["NextToken"] = response["NextToken"]
        else:
            break

    print("\n".join(sorted(secrets)))



#!/usr/bin/env python
import time
import uuid
from pathlib import Path

import boto3
import botocore  # noqa
import cfn_flip.yaml_dumper
import click
import yaml
from cfn_tools import load_yaml

from commands.check_cloudformation import get_lint_result
from commands.dns_cli import get_load_balancer_domain_and_configuration

from .utils import check_aws_conn
from .utils import check_response
from .utils import ensure_cwd_is_repo_root

# This may need to change, once we determine what the default WAF name will be.
WAF_DEFAULT_NAME = "default"


def check_waf(project_session: boto3.Session) -> str:
    click.secho("Checking WAF exists", fg="cyan")
    client = project_session.client("wafv2")

    response = client.list_web_acls(Scope="REGIONAL")
    arn = ""
    for waf in response["WebACLs"]:
        if waf["Name"] == WAF_DEFAULT_NAME:
            arn = waf["ARN"]
    return arn


@click.group()
def waf():
    pass


@waf.command()
@click.option("--app", help="Application Name", required=True)
@click.option(
    "--project-profile", help="aws account profile name for application account", required=True
)
@click.option("--svc", help="Service Name", required=True)
@click.option("--env", help="Environment", required=True)
def attach_waf(app, project_profile, svc, env):
    """Attach default WAF rule to ECS Load Balancer."""

    project_session = check_aws_conn(project_profile)
    waf_arn = check_waf(project_session)
    if not waf_arn:
        click.secho(
            "Default WAF rule does not exists in this AWS account, "
            "please have this created by the SRE team",
            fg="red",
        )
        exit()

    domain_name, load_balancer_configuration = get_load_balancer_domain_and_configuration(
        project_session, app, svc, env
    )

    elb_arn = load_balancer_configuration["LoadBalancerArn"]
    elb_name = load_balancer_configuration["DNSName"]
    waf_client = project_session.client("wafv2")

    response = waf_client.associate_web_acl(WebACLArn=waf_arn, ResourceArn=elb_arn)
    check_response(response)
    click.echo(
        click.style("Default WAF is now associated with ", fg="green")
        + click.style(f"{elb_name} ", fg="white", bold=True)
        + click.style("for domain ", fg="green")
        + click.style(f"{domain_name}", fg="white", bold=True),
    )


def create_stack(cf_client, app, svc, env, raw):
    return cf_client.create_stack(
        StackName=f"{app}-{svc}-{env}-CustomWAFStack",
        TemplateBody=raw,
        TimeoutInMinutes=5,
        OnFailure="DELETE",
        Tags=[
            {"Key": "Name", "Value": "copilot tools custom waf"},
        ],
        ClientRequestToken=uuid.uuid4().hex,
        EnableTerminationProtection=False,
    )


@waf.command()
@click.option("--app", help="Application Name", required=True)
@click.option(
    "--project-profile", help="aws account profile name for application account", required=True
)
@click.option("--svc", help="Service Name", required=True)
@click.option("--env", help="Environment", required=True)
@click.option("--waf-path", help="path to waf.yml file", required=True)
def custom_waf(app, project_profile, svc, env, waf_path):
    """Attach custom WAF to ECS Load Balancer."""

    project_session = check_aws_conn(project_profile)
    ensure_cwd_is_repo_root()
    path = Path().resolve() / waf_path

    # The YAML data_dict needs verification, need to create a function to lint/check YAML formatting.
    result = get_lint_result(str(path))
    if result.returncode != 0:
        click.secho(f"File failed lint check.\n{str(path)}", fg="red")
        exit()

    try:
        with open(path, "r") as fd:
            data_dict = load_yaml(fd)
    except FileNotFoundError:
        click.secho(f"File not found...\n{path}", fg="red")
        exit()

    # dumper is used to resolve the shorthand in YAML file, eg !GetAtt
    dumper = cfn_flip.yaml_dumper.get_dumper(clean_up=True, long_form=False)
    raw = yaml.dump(data_dict, Dumper=dumper, default_flow_style=False, allow_unicode=True)

    cf_client = project_session.client("cloudformation")

    try:
        cs_response = create_stack(cf_client, app, svc, env, raw)
    except cf_client.exceptions.AlreadyExistsException:
        click.echo(
            click.style(
                "CloudFormation Stack already exists, please delete the stack first.\n", fg="red"
            )
            + click.style(f"{app}-{svc}-{env}-CustomWAFStack", fg="red"),
        )
        exit()

    check_response(cs_response)

    response = cf_client.describe_stacks(StackName=cs_response["StackId"])
    while response["Stacks"][0]["StackStatus"] == "CREATE_IN_PROGRESS":
        click.secho("Waiting for WAF resource to be created...", fg="yellow")
        time.sleep(2)
        response = cf_client.describe_stacks(StackName=cs_response["StackId"])

    if response["Stacks"][0]["StackStatus"] == "DELETE_IN_PROGRESS":
        click.secho(
            "Failed to create CloudFormation stack, see AWS webconsole for details", fg="red"
        )
        exit()

    waf_arn = response["Stacks"][0]["Outputs"][0]["OutputValue"]
    click.echo(
        click.style("WAF created: ", fg="yellow") + click.style(f"{waf_arn}", fg="white", bold=True)
    )

    domain_name, load_balancer_configuration = get_load_balancer_domain_and_configuration(
        project_session, app, svc, env
    )
    elb_arn = load_balancer_configuration["LoadBalancerArn"]
    elb_name = load_balancer_configuration["DNSName"]
    waf_client = project_session.client("wafv2")

    # It takes a few seconds for the WAF to be available to use.
    while True:
        try:
            response = waf_client.associate_web_acl(WebACLArn=waf_arn, ResourceArn=elb_arn)
        except waf_client.exceptions.WAFUnavailableEntityException:
            click.secho("Waiting for WAF resource to be available...", fg="yellow")
            time.sleep(2)
            continue
        break

    check_response(response)
    click.echo(
        click.style("Custom WAF is now associated with ", fg="green")
        + click.style(f"{elb_name} ", fg="white", bold=True)
        + click.style("for domain ", fg="green")
        + click.style(f"{domain_name}", fg="white", bold=True),
    )


if __name__ == "__main__":
    waf()



#!/usr/bin/env python
import json
import os
from importlib.metadata import version

import click
from boto3.session import Session
from mypy_boto3_codebuild.client import CodeBuildClient

from .utils import check_aws_conn
from .utils import check_response

AWS_REGION = "eu-west-2"
DEFAULT_CI_BUILDER = "public.ecr.aws/uktrade/ci-image-builder"


def import_pat(pat: str, client: CodeBuildClient):
    response = client.import_source_credentials(
        token=pat,
        serverType="GITHUB",
        authType="PERSONAL_ACCESS_TOKEN",
        shouldOverwrite=True,
    )
    check_response(response)
    click.secho("PAT successfully added", fg="green")


def check_github_conn(client: CodeBuildClient):
    response = client.list_source_credentials()

    # If there are no source code creds defined then AWS is not linked to Github
    if not response["sourceCredentialsInfos"]:
        if not click.confirm(
            click.style(
                "GitHub is not linked in this AWS account\nDo you want to link with a PAT?",
                fg="yellow",
            )
        ):
            exit()

        pat = input(
            """
            Create a bots PAT with the following scope:
                repo:   status: Grants read/write access to public and private repository commit statuses.
                admin:  repo_hook: Grants full control of repository hooks.

            Enter in the PAT here:
            """,
        )

        import_pat(pat, client)


def check_service_role(role_name, project_session: Session) -> str:
    client = project_session.client("iam", region_name=AWS_REGION)

    try:
        response = client.get_role(RoleName=role_name)
        role_arn = response["Role"]["Arn"]

    except client.exceptions.NoSuchEntityException:
        click.echo(
            click.style("Service Role", fg="yellow")
            + click.style(f" {role_name} ", fg="white", bold=True)
            + click.style("does not exist; run: ", fg="yellow")
            + click.style(
                "copilot-helper.py codebuild create-codedeploy-role --type <ci/custom>", fg="cyan"
            )
        )
        role_arn = ""
        exit()

    return role_arn


def update_parameter(project_session: Session, name: str, description: str, value: str):
    client = project_session.client("ssm", region_name=AWS_REGION)

    response = client.put_parameter(
        Name=name,
        Description=description,
        Value=value,
        Type="SecureString",
        Overwrite=True,
        Tier="Standard",
        DataType="text",
    )
    check_response(response)


def check_git_url(git: str) -> str:
    # Ensure the git format is https://github.com/<org>/<repository-name>
    git_part = git.split(":")

    if git_part[0] == "https":
        git_url = git
    elif git_part[0] == "git@github.com":
        git_url = "https://github.com/" + git_part[1]
    else:
        click.echo(
            click.style("Unable to recognise Git URL format, make sure it is either:\n", fg="red")
            + click.style(
                "https://github.com/<org>/<repository-name>\n"
                + "git@github.com:<org>/<repository-name>",
                fg="white",
                bold=True,
            )
        )
        exit()
    return git_url


def modify_project(project_session, update, name, desc, git, branch, buildspec, release, role_type):
    git_url = check_git_url(git)
    role_arn = check_service_role("ci-CodeBuild-role", project_session)
    client = project_session.client("codebuild", region_name=AWS_REGION)
    check_github_conn(client)

    copilot_tools_version = version("dbt-copilot-tools")

    environment = {
        "type": "LINUX_CONTAINER",
        "image": f"{DEFAULT_CI_BUILDER}",
        "computeType": "BUILD_GENERAL1_SMALL",
        "environmentVariables": [
            {
                "name": "COPILOT_TOOLS_VERSION",
                "value": f"{copilot_tools_version}",
                "type": "PLAINTEXT",
            },
        ],
        "privilegedMode": True,
        "imagePullCredentialsType": f"{role_type}",
    }

    source = {
        "type": "GITHUB",
        "location": f"{git_url}",
        "buildspec": f"{buildspec}",
        "auth": {"type": "OAUTH", "resource": "AWS::CodeBuild::SourceCredential"},
    }

    artifacts = {"type": "NO_ARTIFACTS"}

    logsConfig = {
        "cloudWatchLogs": {
            "status": "ENABLED",
        },
    }

    if release:
        pattern = "^refs/tags/.*"
    else:
        pattern = f"^refs/heads/{branch}"

    # Either update project or create a new project
    if update:
        try:
            response = client.update_project(
                name=name,
                description=desc,
                source=source,
                sourceVersion=branch,
                artifacts=artifacts,
                environment=environment,
                serviceRole=role_arn,
                logsConfig=logsConfig,
            )
        except client.exceptions.ResourceNotFoundException:
            click.secho(
                "Unable to update a project that does not exist, remove the --update flag", fg="red"
            )
            exit()

        response_webhook = client.update_webhook(
            projectName=name,
            filterGroups=[
                [
                    {
                        "type": "EVENT",
                        "pattern": "PUSH",
                    },
                    {"type": "HEAD_REF", "pattern": pattern},
                ],
            ],
            buildType="BUILD",
        )

    else:
        try:
            response = client.create_project(
                name=name,
                description=desc,
                source=source,
                sourceVersion=branch,
                artifacts=artifacts,
                environment=environment,
                serviceRole=role_arn,
                logsConfig=logsConfig,
            )

            response_webhook = client.create_webhook(
                projectName=name,
                filterGroups=[
                    [
                        {
                            "type": "EVENT",
                            "pattern": "PUSH",
                        },
                        {"type": "HEAD_REF", "pattern": pattern},
                    ],
                ],
                buildType="BUILD",
            )

        except client.exceptions.ResourceAlreadyExistsException:
            click.secho("Project already exists, use the --update flag", fg="red")
            exit()

    check_response(response)
    check_response(response_webhook)
    click.echo(
        click.style("Codebuild project ", fg="yellow")
        + click.style(f"{name} ", fg="white", bold=True)
        + click.style("updated", fg="yellow")
    )


@click.group()
def codebuild():
    pass


@codebuild.command()
@click.option("--pat", help="PAT Token", required=True)
@click.option("--project-profile", help="aws account profile name", required=True)
def link_github(pat: str, project_profile: str) -> None:
    """Links CodeDeploy to Github via users PAT."""
    project_session = check_aws_conn(project_profile)
    client = project_session.client("codebuild", region_name=AWS_REGION)
    import_pat(pat, client)


@codebuild.command()
@click.option("--project-profile", help="aws account profile name", required=True)
@click.option(
    "--type", type=click.Choice(["ci", "custom"]), help="type of project <ci/custom>", default="ci"
)
def create_codedeploy_role(project_profile: str, type) -> None:
    """Add AWS Role needed for codedeploy."""

    project_session = check_aws_conn(project_profile)
    account_id = project_session.client("sts").get_caller_identity().get("Account")

    current_filepath = os.path.dirname(os.path.realpath(__file__))

    with open(f"{current_filepath}/templates/{type}-codebuild-role-policy.json") as f:
        policy_doc = json.load(f)
    client = project_session.client("iam", region_name=AWS_REGION)

    # A policy must be defined if not present.
    try:
        response = client.create_policy(
            PolicyName=f"{type}-CodeBuild-policy",
            PolicyDocument=json.dumps(policy_doc),
            Description="Custom Policy for codebuild",
            Tags=[
                {"Key": "Name", "Value": "CustomPolicy"},
            ],
        )
        check_response(response)
        click.secho("Policy created", fg="green")

    except client.exceptions.EntityAlreadyExistsException:
        if not click.confirm(click.style("Policy exists.\nDo you want to update it?", fg="yellow")):
            exit()
        try:
            response = client.create_policy_version(
                PolicyArn=f"arn:aws:iam::{account_id}:policy/{type}-CodeBuild-policy",
                PolicyDocument=json.dumps(policy_doc),
                SetAsDefault=True,
            )
            check_response(response)
            click.secho("Policy updated", fg="green")

        except client.exceptions.LimitExceededException:
            click.secho(
                "You have hit the limit of max managed policies, "
                "please delete an existing version and try again",
                fg="red",
            )
            exit()

    with open(f"{current_filepath}/templates/create-codebuild-role.json") as f:
        role_doc = json.load(f)

    # Now create a role if not present and attach policy
    try:
        response = client.create_role(
            RoleName=f"{type}-CodeBuild-role", AssumeRolePolicyDocument=json.dumps(role_doc)
        )
        check_response(response)
        click.secho("Role created", fg="green")
    except client.exceptions.EntityAlreadyExistsException:
        click.secho("Role exists", fg="yellow")

    response = client.attach_role_policy(
        PolicyArn=f"arn:aws:iam::{account_id}:policy/{type}-CodeBuild-policy",
        RoleName=f"{type}-CodeBuild-role",
    )
    check_response(response)
    click.secho("Policy attached to Role", fg="green")


@codebuild.command()
@click.option("--update", is_flag=True, show_default=True, default=False, help="Update config")
@click.option("--name", required=True, help="Name of project")
@click.option("--desc", default="", help="Description of project")
@click.option("--git", required=True, help="Git url of code")
@click.option("--branch", required=True, help="Git branch")
@click.option("--buildspec", required=True, help="Location of buildspec file in repo")
@click.option("--project-profile", required=True, help="aws account profile name")
@click.option(
    "--release",
    is_flag=True,
    show_default=True,
    default=False,
    help="Trigger builds on release tags",
)
def codedeploy(update, name, desc, git, branch, buildspec, project_profile, release):
    """Builds Code build boilerplate."""

    project_session = check_aws_conn(project_profile)
    modify_project(
        project_session,
        update,
        name,
        desc,
        git,
        branch,
        buildspec,
        release,
        "SERVICE_ROLE",
    )


@codebuild.command()
@click.option("--update", is_flag=True, show_default=True, default=False, help="Update config")
@click.option("--name", required=True, help="Name of project")
@click.option("--desc", default="", help="Description of project")
@click.option("--git", required=True, help="Git url of code")
@click.option("--branch", required=True, help="Git branch")
@click.option("--buildspec", required=True, help="Location of buildspec file in repo")
@click.option(
    "--builderimage", default="aws/codebuild/amazonlinux2-x86_64-standard:3.0", help="Builder image"
)
@click.option("--project-profile", required=True, help="aws account profile name")
def buildproject(update, name, desc, git, branch, buildspec, builderimage, project_profile):
    """Builds Code build for ad hoc projects."""

    project_session = check_aws_conn(project_profile)
    modify_project(
        project_session,
        update,
        name,
        desc,
        git,
        branch,
        buildspec,
        builderimage,
        False,
        "CODEBUILD",
    )


@codebuild.command()
@click.option("--name", required=True, help="Name of project")
@click.option("--project-profile", required=True, help="aws account profile name")
def delete_project(name, project_profile):
    """Delete CodeBuild projects."""

    project_session = check_aws_conn(project_profile)
    client = project_session.client("codebuild", region_name=AWS_REGION)

    if not click.confirm(
        click.style("Are you sure you want to delete the project ", fg="yellow")
        + click.style(f"{name}", fg="white", bold=True)
    ):
        exit()

    response = client.delete_project(name=name)
    check_response(response)
    click.secho("Project deleted", fg="green")


@codebuild.command()
@click.option("--workspace", help="Slack Workspace id", required=True)
@click.option("--channel", help="Slack channel id", required=True)
@click.option("--token", help="Slack api token", required=True)
@click.option("--project-profile", help="aws account profile name", required=True)
def slackcreds(workspace, channel, token, project_profile):
    """Add Slack credentials into AWS Parameter Store."""
    project_session = check_aws_conn(project_profile)

    SLACK = {
        "workspace": {
            "name": "/codebuild/slack_workspace_id",
            "description": "Slack Workspace ID",
            "value": workspace,
        },
        "channel": {
            "name": "/codebuild/slack_channel_id",
            "description": "Slack Channel ID",
            "value": channel,
        },
        "token": {
            "name": "/codebuild/slack_api_token",
            "description": "Slack API Token",
            "value": token,
        },
    }

    if not click.confirm(
        click.style(
            "Updating Parameter Store with Slack credentials.\nDo you want to update it?",
            fg="yellow",
        )
    ):
        exit()

    for item, value in SLACK.items():
        update_parameter(project_session, value["name"], value["description"], value["value"])
    click.secho("Paramater Store updated", fg="green")


if __name__ == "__main__":
    codebuild()



#!/usr/bin/env python

import os
import time
from typing import Tuple

import click
import yaml
from boto3 import Session

from .utils import check_aws_conn
from .utils import check_response
from .utils import ensure_cwd_is_repo_root

# To do
# -----
# Check user has logged into the aws accounts before scanning the accounts
# When adding records from parent to subdomain, if ok, it then should remove them from parent domain
# (run a test before removing)

# Base domain depth
MAX_DOMAIN_DEPTH = 2
AWS_CERT_REGION = "eu-west-2"


def wait_for_certificate_validation(acm_client, certificate_arn, sleep_time=10, timeout=600):
    click.secho(f"Waiting up to {timeout} seconds for certificate to be validated...", fg="yellow")
    status = acm_client.describe_certificate(CertificateArn=certificate_arn)["Certificate"][
        "Status"
    ]
    elapsed_time = 0
    while status == "PENDING_VALIDATION":
        if elapsed_time >= timeout:
            raise Exception(
                f"Timeout ({timeout}s) reached for certificate validation, might be worth checking things in the AWS Console"
            )
        click.echo(
            click.style(f"{certificate_arn}", fg="white", bold=True)
            + click.style(
                f": Waiting {sleep_time}s for validation, {elapsed_time}s elapsed, {timeout - elapsed_time}s until we give up...",
                fg="yellow",
            ),
        )
        time.sleep(sleep_time)
        status = acm_client.describe_certificate(CertificateArn=certificate_arn)["Certificate"][
            "Status"
        ]
        elapsed_time += sleep_time

    if status == "ISSUED":
        click.secho("Certificate validated...", fg="green")
    else:
        raise Exception(f"""Certificate validation failed with the status "{status}".""")


def create_cert(client, domain_client, domain, base_len):
    # Check if cert is present.
    arn = ""
    resp = client.list_certificates(
        CertificateStatuses=[
            "PENDING_VALIDATION",
            "ISSUED",
            "INACTIVE",
            "EXPIRED",
            "VALIDATION_TIMED_OUT",
            "REVOKED",
            "FAILED",
        ],
        MaxItems=500,
    )

    # Need to check if cert status is issued, if pending need to update dns
    for cert in resp["CertificateSummaryList"]:
        if domain == cert["DomainName"]:
            click.secho("Certificate already exists, do not need to create.", fg="green")
            return cert["CertificateArn"]

    if not click.confirm(
        click.style("Creating Certificate for ", fg="yellow")
        + click.style(f"{domain}\n", fg="white", bold=True)
        + click.style("Do you want to continue?", fg="yellow"),
    ):
        exit()

    parts = domain.split(".")

    # We only want to create domains max 2 deep so remove all sub domains in excess
    parts_to_remove = len(parts) - base_len - MAX_DOMAIN_DEPTH
    domain_to_create = ".".join(parts[parts_to_remove:]) + "."
    click.secho(domain_to_create, fg="yellow")
    # cert_client = DNSValidatedACMCertClient(domain=domain, profile='intranet')
    response = client.request_certificate(DomainName=domain, ValidationMethod="DNS")

    arn = response["CertificateArn"]

    # Create DNS validation records
    # Need a pause for it to populate the DNS resource records
    response = client.describe_certificate(CertificateArn=arn)
    while (
        response["Certificate"].get("DomainValidationOptions") is None
        or response["Certificate"]["DomainValidationOptions"][0].get("ResourceRecord") is None
    ):
        click.secho("Waiting for DNS records...", fg="yellow")
        time.sleep(2)
        response = client.describe_certificate(CertificateArn=arn)

    cert_record = response["Certificate"]["DomainValidationOptions"][0]["ResourceRecord"]

    click.secho(f"Looking fo ID of domain {domain_to_create}...", fg="yellow")
    domain_id = False
    response = domain_client.list_hosted_zones_by_name()
    for hz in response["HostedZones"]:
        if hz["Name"] == domain_to_create:
            domain_id = hz["Id"]
            break
    if not domain_id:
        # Will got here more than once during manual testing, it might be a race condition we need to handle better
        click.secho(
            f"Unable to find Domain ID for {domain_to_create} in the hosted zones",
            fg="red",
            bold=True,
        )
        exit(1)

    # Add NS records of subdomain to parent
    click.secho(domain_id, fg="yellow")

    if not click.confirm(
        click.style("Updating DNS record for certificate ", fg="yellow")
        + click.style(f"{domain}", fg="white", bold=True)
        + click.style(" with value ", fg="yellow")
        + click.style(f"""{cert_record["Value"]}\n""", fg="white", bold=True)
        + click.style("Do you want to continue?", fg="yellow"),
    ):
        exit()

    response = domain_client.change_resource_record_sets(
        HostedZoneId=domain_id,
        ChangeBatch={
            "Changes": [
                {
                    "Action": "CREATE",
                    "ResourceRecordSet": {
                        "Name": cert_record["Name"],
                        "Type": cert_record["Type"],
                        "TTL": 300,
                        "ResourceRecords": [
                            {"Value": cert_record["Value"]},
                        ],
                    },
                },
            ],
        },
    )

    # Wait for certificate to get to validation state before continuing.
    # Will upped the timeout from 600 to 1200 because he repeatedly saw it timeout at 600
    # and wants to give it a chance to take longer in case that's all that's wrong.
    wait_for_certificate_validation(client, certificate_arn=arn, timeout=1200)

    return arn


def add_records(client, records, subdom_id, action):
    if records["Type"] == "A":
        response = client.change_resource_record_sets(
            HostedZoneId=subdom_id,
            ChangeBatch={
                "Comment": "Record created for copilot",
                "Changes": [
                    {
                        "Action": action,
                        "ResourceRecordSet": {
                            "Name": records["Name"],
                            "Type": records["Type"],
                            "AliasTarget": {
                                "HostedZoneId": records["AliasTarget"]["HostedZoneId"],
                                "DNSName": records["AliasTarget"]["DNSName"],
                                "EvaluateTargetHealth": records["AliasTarget"][
                                    "EvaluateTargetHealth"
                                ],
                            },
                        },
                    },
                ],
            },
        )
    else:
        response = client.change_resource_record_sets(
            HostedZoneId=subdom_id,
            ChangeBatch={
                "Comment": "Record created for copilot",
                "Changes": [
                    {
                        "Action": action,
                        "ResourceRecordSet": {
                            "Name": records["Name"],
                            "Type": records["Type"],
                            "TTL": records["TTL"],
                            "ResourceRecords": [
                                {"Value": records["ResourceRecords"][0]["Value"]},
                            ],
                        },
                    },
                ],
            },
        )

    check_response(response)
    click.echo(
        click.style(f"{records['Name']}, Type: {records['Type']}", fg="white", bold=True)
        + click.style("Added.", fg="magenta"),
    )
    return response["ChangeInfo"]["Status"]


def check_for_records(client, parent_id, subdom, subdom_id):
    records_to_add = []
    response = client.list_resource_record_sets(
        HostedZoneId=parent_id,
    )

    for records in response["ResourceRecordSets"]:
        if subdom in records["Name"]:
            click.secho(records["Name"] + " found", fg="green")
            records_to_add.append(records)
            add_records(client, records, subdom_id, "UPSERT")
    return True


def create_hosted_zone(client, domain, start_domain, base_len):
    parts = domain.split(".")

    # We only want to create domains max 2 deep so remove all sub domains in excess
    parts_to_remove = len(parts) - base_len - MAX_DOMAIN_DEPTH
    domain_to_create = parts[parts_to_remove:]

    # Walk back domain name
    for x in reversed(range(len(domain_to_create) - (len(start_domain.split(".")) - 1))):
        subdom = ".".join(domain_to_create[(x):]) + "."

        if not click.confirm(
            click.style("About to create domain: ", fg="cyan")
            + click.style(f"{subdom}\n", fg="white", bold=True)
            + click.style("Do you want to continue?", fg="cyan"),
        ):
            exit()

        parent = ".".join(subdom.split(".")[1:])
        response = client.list_hosted_zones_by_name()

        for hz in response["HostedZones"]:
            if hz["Name"] == parent:
                parent_id = hz["Id"]
                break

        click.secho(f"Creating hosted zone for {subdom}...", fg="yellow")
        response = client.create_hosted_zone(
            Name=subdom,
            # Timestamp is on the end because CallerReference must be unique for every call
            CallerReference=f"{subdom}_from_code_{int(time.time())}",
        )
        ns_records = response["DelegationSet"]
        subdom_id = response["HostedZone"]["Id"]

        # Check if records existed in the parent domain, if so they need to be copied to sub domain.
        check_for_records(client, parent_id, subdom, subdom_id)

        if not click.confirm(
            click.style(f"Updating parent {parent} domain with records: ", fg="cyan")
            + click.style(f"{ns_records['NameServers']}\n", fg="white", bold=True)
            + click.style("Do you want to continue?", fg="cyan"),
        ):
            exit()

        # Add NS records of subdomain to parent
        nameservers = ns_records["NameServers"]
        # append  . to make fqdn
        nameservers = [f"{nameserver}." for nameserver in nameservers]
        nameserver_resource_records = [{"Value": nameserver} for nameserver in nameservers]

        client.change_resource_record_sets(
            HostedZoneId=parent_id,
            ChangeBatch={
                "Changes": [
                    {
                        "Action": "CREATE",
                        "ResourceRecordSet": {
                            "Name": subdom,
                            "Type": "NS",
                            "TTL": 300,
                            "ResourceRecords": nameserver_resource_records,
                        },
                    },
                ],
            },
        )

    return True


def check_r53(domain_session, project_session, domain, base_domain):
    # find the hosted zone
    domain_client = domain_session.client("route53")
    acm_client = project_session.client("acm", region_name=AWS_CERT_REGION)

    # create the certificate
    response = domain_client.list_hosted_zones_by_name()

    hosted_zones = {}
    for hz in response["HostedZones"]:
        hosted_zones[hz["Name"]] = hz

    # Check if base domain is valid
    if base_domain[-1] != ".":
        base_domain = base_domain + "."

    if base_domain not in hosted_zones:
        click.secho(
            f"The base domain: {base_domain} does not exist in your AWS domain account \
                {response['HostedZones']}",
            fg="red",
        )
        exit()

    base_len = len(base_domain.split(".")) - 1
    parts = domain.split(".")

    for _ in range(len(parts) - 1):
        subdom = ".".join(parts) + "."
        click.secho(f"Searching for {subdom}... ", fg="yellow")
        if subdom in hosted_zones:
            click.secho("Found hosted zone " + hosted_zones[subdom]["Name"], fg="green")

            # We only want to go 2 sub domains deep in R53
            if (len(parts) - base_len) < MAX_DOMAIN_DEPTH:
                click.secho("Creating Hosted Zone", fg="magenta")
                create_hosted_zone(domain_client, domain, subdom, base_len)

            break

        parts.pop(0)
    else:
        # This should only occur when base domain this needs is not found
        click.secho(f"Root Domain not found for {domain}", fg="red")
        return

    # add records to hosted zone to validate certificate
    cert_arn = create_cert(acm_client, domain_client, domain, base_len)

    return cert_arn


def get_load_balancer_domain_and_configuration(
    project_session: Session, app: str, svc: str, env: str
) -> Tuple[str, dict]:
    def separate_hyphenated_application_environment_and_service(
        hyphenated_string, number_of_items_of_interest, number_of_trailing_items
    ):
        # The application name may be hyphenated, so we start splitting
        # at the hyphen after the first item of interest and return the
        # items of interest only...
        return hyphenated_string.rsplit(
            "-", number_of_trailing_items + number_of_items_of_interest - 1
        )[:number_of_items_of_interest]

    proj_client = project_session.client("ecs")

    response = proj_client.list_clusters()
    check_response(response)
    no_items = True
    for cluster_arn in response["clusterArns"]:
        cluster_name = cluster_arn.split("/")[1]
        cluster_app, cluster_env = separate_hyphenated_application_environment_and_service(
            cluster_name, 2, 2
        )
        if cluster_app == app and cluster_env == env:
            no_items = False
            break

    if no_items:
        click.echo(
            click.style("There are no clusters matching ", fg="red")
            + click.style(f"{app} ", fg="white", bold=True)
            + click.style("in this aws account", fg="red"),
        )
        exit()

    response = proj_client.list_services(cluster=cluster_name)
    check_response(response)
    no_items = True
    for service_arn in response["serviceArns"]:
        fully_qualified_service_name = service_arn.split("/")[2]
        (
            service_app,
            service_env,
            service_name,
        ) = separate_hyphenated_application_environment_and_service(
            fully_qualified_service_name, 3, 2
        )
        if service_app == app and service_env == env and service_name == svc:
            no_items = False
            break

    if no_items:
        click.echo(
            click.style("There are no services matching ", fg="red")
            + click.style(f"{svc}", fg="white", bold=True)
            + click.style(" in this aws account", fg="red"),
        )
        exit()

    elb_client = project_session.client("elbv2")

    elb_arn = elb_client.describe_target_groups(
        TargetGroupArns=[
            proj_client.describe_services(
                cluster=cluster_name,
                services=[
                    fully_qualified_service_name,
                ],
            )["services"][0]["loadBalancers"][0]["targetGroupArn"],
        ],
    )["TargetGroups"][0]["LoadBalancerArns"][0]

    response = elb_client.describe_load_balancers(LoadBalancerArns=[elb_arn])
    check_response(response)

    # Find the domain name
    with open(f"./copilot/{svc}/manifest.yml", "r") as fd:
        conf = yaml.safe_load(fd)
        if "environments" in conf:
            for domain in conf["environments"].items():
                if domain[0] == env:
                    domain_name = domain[1]["http"]["alias"]

        # What happens if domain_name isn't set? Should we raise an error? Return default ? Or None?

    return domain_name, response["LoadBalancers"][0]


@click.group()
def domain():
    pass


@domain.command()
@click.option(
    "--domain-profile", help="aws account profile name for R53 domains account", required=True
)
@click.option(
    "--project-profile", help="aws account profile name for certificates account", required=True
)
@click.option("--base-domain", help="root domain", required=True)
def check_domain(domain_profile, project_profile, base_domain):
    """Scans to see if Domain exists."""

    # If you need to reset to debug this command, you will need to delete any of the following
    # which have been created:
    # the certificate in your application's AWS account,
    # the hosted zone for the application environment in the dev AWS account,
    # and the applications records on the hosted zone for the environment in the dev AWS account.

    path = "copilot"

    domain_session = check_aws_conn(domain_profile)
    project_session = check_aws_conn(project_profile)

    if not os.path.exists(path):
        click.secho("Please check path, manifest file not found", fg="red")
        exit()

    if path.split(".")[-1] == "yml" or path.split(".")[-1] == "yaml":
        click.secho("Please do not include the filename in the path", fg="red")
        exit()

    cert_list = {}
    for root, dirs, files in os.walk(path):
        for file in files:
            if file == "manifest.yml" or file == "manifest.yaml":
                # Need to check that the manifest file is correctly configured.
                with open(os.path.join(root, file), "r") as fd:
                    conf = yaml.safe_load(fd)
                    if "environments" in conf:
                        click.echo(
                            click.style("Checking file: ", fg="cyan")
                            + click.style(os.path.join(root, file), fg="white"),
                        )
                        click.secho("Domains listed in manifest file", fg="cyan", underline=True)

                        for env, domain in conf["environments"].items():
                            click.secho(
                                "\nEnvironment: " + env + " => Domain: " + domain["http"]["alias"],
                                fg="yellow",
                                bold=True,
                            )
                            cert_arn = check_r53(
                                domain_session,
                                project_session,
                                domain["http"]["alias"],
                                base_domain,
                            )
                            cert_list.update({domain["http"]["alias"]: cert_arn})

    if cert_list:
        click.secho("\nHere are your Certificate ARNs:", fg="cyan")
        for domain, cert in cert_list.items():
            click.secho(f"Domain: {domain}\t => Cert ARN: {cert}", fg="white", bold=True)
    else:
        click.secho("No domains found, please check the manifest file", fg="red")


@domain.command()
@click.option("--app", help="Application Name", required=True)
@click.option(
    "--domain-profile", help="aws account profile name for R53 domains account", required=True
)
@click.option(
    "--project-profile", help="aws account profile name for application account", required=True
)
@click.option("--svc", help="Service Name", required=True)
@click.option("--env", help="Environment", required=True)
def assign_domain(app, domain_profile, project_profile, svc, env):
    """Check R53 domain is pointing to the correct ECS Load Blanacer."""
    domain_session = check_aws_conn(domain_profile)
    project_session = check_aws_conn(project_profile)

    ensure_cwd_is_repo_root()

    # Find the Load Balancer name.
    domain_name, load_balancer_configuration = get_load_balancer_domain_and_configuration(
        project_session, app, svc, env
    )
    elb_name = load_balancer_configuration["DNSName"]

    click.echo(
        click.style("The Domain: ", fg="yellow")
        + click.style(f"{domain_name}\n", fg="white", bold=True)
        + click.style("has been assigned the Load Balancer: ", fg="yellow")
        + click.style(f"{elb_name}\n", fg="white", bold=True)
        + click.style("Checking to see if this is in R53", fg="yellow"),
    )

    domain_client = domain_session.client("route53")
    response = domain_client.list_hosted_zones_by_name()
    check_response(response)

    # Scan R53 Zone for matching domains and update records if needed.
    hosted_zones = {}
    for hz in response["HostedZones"]:
        hosted_zones[hz["Name"]] = hz

    parts = domain_name.split(".")
    for _ in range(len(parts) - 1):
        subdom = ".".join(parts) + "."
        click.echo(
            click.style("Searching for ", fg="yellow")
            + click.style(f"{subdom}..", fg="white", bold=True)
        )

        if subdom in hosted_zones:
            click.echo(
                click.style("Found hosted zone ", fg="yellow")
                + click.style(hosted_zones[subdom]["Name"], fg="white", bold=True),
            )
            hosted_zone_id = hosted_zones[subdom]["Id"]

            # Does record existing
            response = domain_client.list_resource_record_sets(
                HostedZoneId=hosted_zone_id,
            )
            check_response(response)

            for record in response["ResourceRecordSets"]:
                if domain_name == record["Name"][:-1]:
                    click.echo(
                        click.style("Record: ", fg="yellow")
                        + click.style(f"{record['Name']} found", fg="white", bold=True),
                    )
                    click.echo(
                        click.style("is pointing to LB ", fg="yellow")
                        + click.style(
                            f"{record['ResourceRecords'][0]['Value']}", fg="white", bold=True
                        ),
                    )
                    if record["ResourceRecords"][0]["Value"] != elb_name:
                        if click.confirm(
                            click.style("This doesnt match with the current LB ", fg="yellow")
                            + click.style(f"{elb_name}", fg="white", bold=True)
                            + click.style("Do you wish to update the record?", fg="yellow"),
                        ):
                            record = {
                                "Name": domain_name,
                                "Type": "CNAME",
                                "TTL": 300,
                                "ResourceRecords": [{"Value": elb_name}],
                            }
                            add_records(domain_client, record, hosted_zone_id, "UPSERT")
                    else:
                        click.secho("No need to add as it already exists", fg="green")
                    exit()

            record = {
                "Name": domain_name,
                "Type": "CNAME",
                "TTL": 300,
                "ResourceRecords": [{"Value": elb_name}],
            }

            if not click.confirm(
                click.style("Creating R53 record: ", fg="yellow")
                + click.style(
                    f"{record['Name']} -> {record['ResourceRecords'][0]['Value']}\n",
                    fg="white",
                    bold=True,
                )
                + click.style("In Domain: ", fg="yellow")
                + click.style(f"{subdom}", fg="white", bold=True)
                + click.style("\tZone ID: ", fg="yellow")
                + click.style(f"{hosted_zone_id}\n", fg="white", bold=True)
                + click.style("Do you want to continue?", fg="yellow"),
            ):
                exit()
            add_records(domain_client, record, hosted_zone_id, "CREATE")
            exit()

        parts.pop(0)

    else:
        click.echo(
            click.style("No hosted zone found for ", fg="yellow")
            + click.style(f"{domain_name}", fg="white", bold=True),
        )
        return


if __name__ == "__main__":
    domain()



#!/usr/bin/env python
import os
from pathlib import Path
from shutil import rmtree
from subprocess import run

import click

from commands.bootstrap_cli import make_config
from commands.copilot_cli import make_addons

BASE_DIR = Path(__file__).parent.parent


@click.group(invoke_without_command=True, chain=True)
@click.pass_context
def check_cloudformation(ctx: click.Context) -> None:
    """
    Runs the checks passed in the command arguments.

    If no argument is passed, it will run all the checks.
    """

    ctx.obj = {"passing_checks": [], "failing_checks": []}

    prepare_cloudformation_templates(ctx)

    if ctx.invoked_subcommand is None:
        click.secho(f"\n>>> Running all checks", fg="yellow")
        for name, command in ctx.command.commands.items():
            ctx.invoke(command)


def get_lint_result(path: str):
    command = [
        "cfn-lint",
        path,
        "--ignore-templates",
        # addons.parameters.yml is not a CloudFormation template file
        f"{BASE_DIR}/tests/test-application/copilot/**/addons/addons.parameters.yml",
        # "W2001 Parameter Env not used" is ignored becomes Copilot addons require
        # parameters even if they are not used in the Cloudformation template.
        "--ignore-checks",
        "W2001",
    ]

    click.secho(f"\n>>> Running lint check", fg="yellow")
    click.secho(f"""    {" ".join(command)}\n""", fg="yellow")

    return run(command, capture_output=True)


@check_cloudformation.command()
@click.pass_context
def lint(ctx: click.Context) -> None:
    """Runs cfn-lint against the generated CloudFormation templates."""

    BASE_DIR = Path(__file__).parent.parent

    result = get_lint_result(f"{BASE_DIR}/tests/test-application/copilot/**/addons/*.yml")

    click.secho(result.stdout.decode())
    if result.returncode == 0:
        ctx.obj["passing_checks"].append("lint")
    else:
        click.secho(result.stderr.decode())
        ctx.obj["failing_checks"].append("lint")


@check_cloudformation.result_callback()
@click.pass_context
def process_result(ctx: click.Context, result) -> None:
    if ctx.obj["passing_checks"]:
        click.secho("\nThe CloudFormation templates passed the following checks :-)", fg="green")
        for passing_check in ctx.obj["passing_checks"]:
            click.secho(f"  - {passing_check}", fg="green")

    if ctx.obj["failing_checks"]:
        click.secho("\nThe CloudFormation templates failed the following checks :-(", fg="red")
        for failing_check in ctx.obj["failing_checks"]:
            click.secho(f"  - {failing_check}", fg="red")
        exit(1)


def prepare_cloudformation_templates(ctx: click.Context) -> None:
    click.secho(f"\n>>> Preparing CloudFormation templates\n", fg="yellow")
    os.chdir(f"{BASE_DIR}/tests/test-application")
    copilot_directory = Path("./copilot")
    if copilot_directory.exists():
        rmtree(copilot_directory)
    ctx.invoke(make_config)
    ctx.invoke(make_addons)



#!/usr/bin/env python

from importlib.metadata import version

import click

from commands.bootstrap_cli import bootstrap as bootstrap_commands
from commands.check_cloudformation import (
    check_cloudformation as check_cloudformation_command,
)
from commands.codebuild_cli import codebuild as codebuild_commands
from commands.conduit_cli import conduit as conduit_commands
from commands.copilot_cli import copilot as copilot_commands
from commands.dns_cli import domain as domain_commands
from commands.waf_cli import waf as waf_commands


@click.group()
@click.version_option(
    version=version("dbt-copilot-tools"),
    message=f"dbt-copilot-tools %(version)s",
)
def copilot_helper():
    pass


copilot_helper.add_command(bootstrap_commands)
copilot_helper.add_command(check_cloudformation_command)
copilot_helper.add_command(codebuild_commands)
copilot_helper.add_command(conduit_commands)
copilot_helper.add_command(copilot_commands)
copilot_helper.add_command(domain_commands)
copilot_helper.add_command(waf_commands)

if __name__ == "__main__":
    copilot_helper()



import re
import subprocess
import time

import boto3
import click


class ConduitError(Exception):
    pass


class InvalidAddonTypeConduitError(ConduitError):
    pass


class NoClusterConduitError(ConduitError):
    pass


class SecretNotFoundConduitError(ConduitError):
    pass


class CreateTaskTimeoutConduitError(ConduitError):
    pass


CONDUIT_DOCKER_IMAGE_LOCATION = "public.ecr.aws/uktrade/tunnel"
CONDUIT_ADDON_TYPES = [
    "opensearch",
    "postgres",
    "redis",
]


def normalise_string(to_normalise: str) -> str:
    output = re.sub("[^0-9a-zA-Z]+", "-", to_normalise)
    return output.lower()


def get_cluster_arn(app: str, env: str) -> str:
    ecs_client = boto3.client("ecs")

    for cluster_arn in ecs_client.list_clusters()["clusterArns"]:
        tags_response = ecs_client.list_tags_for_resource(resourceArn=cluster_arn)
        tags = tags_response["tags"]

        app_key_found = False
        env_key_found = False
        cluster_key_found = False

        for tag in tags:
            if tag["key"] == "copilot-application" and tag["value"] == app:
                app_key_found = True
            if tag["key"] == "copilot-environment" and tag["value"] == env:
                env_key_found = True
            if tag["key"] == "aws:cloudformation:logical-id" and tag["value"] == "Cluster":
                cluster_key_found = True

        if app_key_found and env_key_found and cluster_key_found:
            return cluster_arn

    raise NoClusterConduitError


def get_connection_secret_arn(app: str, env: str, name: str) -> str:
    connection_secret_id = f"/copilot/{app}/{env}/secrets/{name}"

    secrets_manager = boto3.client("secretsmanager")
    ssm = boto3.client("ssm")

    try:
        return secrets_manager.describe_secret(SecretId=connection_secret_id)["ARN"]
    except secrets_manager.exceptions.ResourceNotFoundException:
        pass

    try:
        return ssm.get_parameter(Name=connection_secret_id, WithDecryption=False)["Parameter"][
            "ARN"
        ]
    except ssm.exceptions.ParameterNotFound:
        pass

    raise SecretNotFoundConduitError(name)


def create_addon_client_task(app: str, env: str, addon_type: str, addon_name: str):
    connection_secret_arn = get_connection_secret_arn(app, env, addon_name.upper())

    subprocess.call(
        f"copilot task run --app {app} --env {env} "
        f"--task-group-name conduit-{app}-{env}-{normalise_string(addon_name)} "
        f"--image {CONDUIT_DOCKER_IMAGE_LOCATION}:{addon_type} "
        f"--secrets CONNECTION_SECRET={connection_secret_arn} "
        "--platform-os linux "
        "--platform-arch arm64",
        shell=True,
    )


def addon_client_is_running(app: str, env: str, cluster_arn: str, addon_name: str) -> bool:
    tasks = boto3.client("ecs").list_tasks(
        cluster=cluster_arn,
        desiredStatus="RUNNING",
        family=f"copilot-conduit-{app}-{env}-{normalise_string(addon_name)}",
    )

    if not tasks["taskArns"]:
        return False

    described_tasks = boto3.client("ecs").describe_tasks(
        cluster=cluster_arn, tasks=tasks["taskArns"]
    )

    # The ExecuteCommandAgent often takes longer to start running than the task and without the
    # agent it's not possible to exec into a task.
    for task in described_tasks["tasks"]:
        for container in task["containers"]:
            for agent in container["managedAgents"]:
                if agent["name"] == "ExecuteCommandAgent" and agent["lastStatus"] == "RUNNING":
                    return True

    return False


def connect_to_addon_client_task(app: str, env: str, cluster_arn: str, addon_name: str):
    tries = 0
    running = False

    while tries < 15 and not running:
        tries += 1

        if addon_client_is_running(app, env, cluster_arn, addon_name):
            running = True
            subprocess.call(
                "copilot task exec "
                f"--app {app} --env {env} "
                f"--name conduit-{app}-{env}-{normalise_string(addon_name)} "
                f"--command bash",
                shell=True,
            )

        time.sleep(1)

    if not running:
        raise CreateTaskTimeoutConduitError


def start_conduit(app: str, env: str, addon_type: str, addon_name: str = None):
    if addon_type not in CONDUIT_ADDON_TYPES:
        raise InvalidAddonTypeConduitError(addon_type)

    cluster_arn = get_cluster_arn(app, env)
    addon_name = addon_name or addon_type

    if not addon_client_is_running(app, env, cluster_arn, addon_name):
        create_addon_client_task(app, env, addon_type, addon_name)
    connect_to_addon_client_task(app, env, cluster_arn, addon_name)


@click.command()
@click.argument("addon_type")
@click.option("--app", help="AWS application name", required=True)
@click.option("--env", help="AWS environment name", required=True)
@click.option("--addon-name", help="Name of custom addon", required=False)
def conduit(addon_type: str, app: str, env: str, addon_name: str):
    """Create a conduit connection to a ADDON_TYPE backing service."""
    try:
        start_conduit(app, env, addon_type, addon_name)
    except InvalidAddonTypeConduitError:
        click.secho(
            f"""Addon type "{addon_type}" does not exist, try one of {", ".join(CONDUIT_ADDON_TYPES)}.""",
            fg="red",
        )
        exit(1)
    except NoClusterConduitError:
        click.secho(f"""No ECS cluster found for "{app}" in "{env}" environment.""", fg="red")
        exit(1)
    except SecretNotFoundConduitError as err:
        click.secho(
            f"""No secret called "{addon_name or err}" for "{app}" in "{env}" environment.""",
            fg="red",
        )
        exit(1)
    except CreateTaskTimeoutConduitError:
        click.secho(
            f"""Client ({addon_type}) ECS task has failed to start for "{app}" in "{env}" environment.""",
            fg="red",
        )
        exit(1)



import re
from pathlib import Path

import boto3
import botocore
import click
import jinja2

from commands.exceptions import ValidationException
from commands.jinja2_tags import VersionTag

SSM_BASE_PATH = "/copilot/{app}/{env}/secrets/"
SSM_PATH = "/copilot/{app}/{env}/secrets/{name}"


def mkdir(base, path):
    if (base / path).exists():
        return f"Directory {path} exists; doing nothing"

    (base / path).mkdir(parents=True)
    return f"Directory {path} created"


def mkfile(base, path, contents, overwrite=False):
    file_exists = (base / path).exists()

    if file_exists and not overwrite:
        return f"File {path} exists; doing nothing"

    action = "overwritten" if overwrite else "created"

    with open(base / path, "w") as fd:
        fd.write(contents)

    return f"File {path} {action}"


def camel_case(s):
    s = re.sub(r"(_|-)+", " ", s).title().replace(" ", "")
    return "".join([s[0].lower(), s[1:]])


def set_ssm_param(
    app, env, param_name, param_value, overwrite, exists, description="Copied from Cloud Foundry."
):
    client = boto3.client("ssm")

    args = dict(
        Name=param_name,
        Description=description,
        Value=param_value,
        Type="SecureString",
        Overwrite=overwrite,
        Tags=[
            {"Key": "copilot-application", "Value": app},
            {"Key": "copilot-environment", "Value": env},
        ],
        Tier="Intelligent-Tiering",
    )

    if overwrite and not exists:
        raise ValidationException(
            """Arguments "overwrite" is set to True, but "exists" is set to False."""
        )

    if overwrite and exists:
        # Tags can't be updated when overwriting
        del args["Tags"]

    client.put_parameter(**args)


def get_ssm_secret_names(app, env):
    client = boto3.client("ssm")

    path = SSM_BASE_PATH.format(app=app, env=env)

    params = dict(
        Path=path,
        Recursive=False,
        WithDecryption=True,
        MaxResults=10,
    )

    secret_names = []

    while True:
        response = client.get_parameters_by_path(**params)

        for secret in response["Parameters"]:
            secret_names.append(secret["Name"])

        if "NextToken" in response:
            params["NextToken"] = response["NextToken"]
        else:
            break

    return sorted(secret_names)


def get_ssm_secrets(app, env):
    """Return secrets from AWS Parameter Store as a list of tuples with the
    secret name and secret value."""

    client = boto3.client("ssm")

    path = SSM_BASE_PATH.format(app=app, env=env)

    params = dict(
        Path=path,
        Recursive=False,
        WithDecryption=True,
        MaxResults=10,
    )

    secrets = []

    while True:
        response = client.get_parameters_by_path(**params)

        for secret in response["Parameters"]:
            secrets.append((secret["Name"], secret["Value"]))

        if "NextToken" in response:
            params["NextToken"] = response["NextToken"]
        else:
            break

    return sorted(secrets)


def setup_templates():
    Path(__file__).parent.parent / Path("templates")
    templateLoader = jinja2.PackageLoader("commands")
    templateEnv = jinja2.Environment(loader=templateLoader, keep_trailing_newline=True)
    templateEnv.add_extension(VersionTag)

    return templateEnv


def ensure_cwd_is_repo_root():
    """Exit if we're not in the root of the repo."""

    if not Path("./copilot").exists() or not Path("./copilot").is_dir():
        click.secho(
            "Cannot find copilot directory. Run this command in the root of the deployment repository.",
            bg="red",
        )
        exit(1)


def check_aws_conn(aws_profile: str) -> boto3.session.Session:
    # Check that the aws profile exists and is set.
    click.secho(f"""Checking AWS connection for profile "{aws_profile}"...""", fg="cyan")

    try:
        session = boto3.session.Session(profile_name=aws_profile)
    except botocore.exceptions.ProfileNotFound:
        click.secho(f"""AWS profile "{aws_profile}" is not configured.""", fg="red")
        exit()
    except botocore.errorfactory.UnauthorizedException:
        click.secho(
            "The SSO session associated with this profile has expired or is otherwise invalid.  "
            "To refresh this SSO session run aws sso login with the corresponding profile",
            fg="red",
        )
        exit()

    sts = session.client("sts")
    try:
        sts.get_caller_identity()
        click.secho("Credentials are valid.", fg="green")
    except botocore.exceptions.SSOTokenLoadError:
        click.secho(
            f"Credentials are NOT valid.  \nPlease login with: aws sso login --profile {aws_profile}",
            fg="red",
        )
        exit()
    except botocore.exceptions.UnauthorizedSSOTokenError:
        click.secho(
            "The SSO session associated with this profile has expired or is otherwise invalid.  "
            "To refresh this SSO session run aws sso login with the corresponding profile",
            fg="red",
        )
        exit()

    alias_client = session.client("iam")
    account_name = alias_client.list_account_aliases()["AccountAliases"]
    if account_name:
        click.echo(
            click.style("Logged in with AWS account: ", fg="yellow")
            + click.style(
                f"{account_name[0]}/{sts.get_caller_identity()['Account']}", fg="white", bold=True
            ),
        )
    else:
        click.echo(
            click.style("Logged in with AWS account id: ", fg="yellow")
            + click.style(f"{sts.get_caller_identity()['Account']}", fg="white", bold=True),
        )
    click.echo(
        click.style("User: ", fg="yellow")
        + click.style(
            f"{(sts.get_caller_identity()['UserId']).split(':')[-1]}\n", fg="white", bold=True
        ),
    )

    return session


def check_response(response):
    if response["ResponseMetadata"]["HTTPStatusCode"] != 200:
        click.secho(
            f"Unknown response error from AWS.\nStatus Code: {response['ResponseMetadata']['HTTPStatusCode']}",
            fg="red",
        )
        exit()



#!/usr/bin/env python

from collections import defaultdict
from pathlib import Path

import click
import yaml
from botocore.exceptions import ClientError
from cloudfoundry_client.client import CloudFoundryClient
from schema import Optional
from schema import Schema

from commands.utils import SSM_PATH
from commands.utils import check_aws_conn
from commands.utils import get_ssm_secret_names
from commands.utils import get_ssm_secrets
from commands.utils import mkdir
from commands.utils import mkfile
from commands.utils import set_ssm_param
from commands.utils import setup_templates

config_schema = Schema(
    {
        "app": str,
        "environments": {str: {Optional("certificate_arns"): [str]}},
        "services": [
            {
                "name": str,
                "type": lambda s: s
                in (
                    "public",
                    "backend",
                ),
                "repo": str,
                "image_location": str,
                Optional("notes"): str,
                Optional("secrets_from"): str,
                "environments": {
                    str: {
                        "paas": str,
                        Optional("url"): str,
                        Optional("ipfilter"): bool,
                    },
                },
                Optional("backing-services"): [
                    {
                        "name": str,
                        "type": lambda s: s
                        in (
                            "s3",
                            "s3-policy",
                            "aurora-postgres",
                            "rds-postgres",
                            "redis",
                            "opensearch",
                        ),
                        Optional("paas-description"): str,
                        Optional("paas-instance"): str,
                        Optional("notes"): str,
                        Optional("bucket_name"): str,  # for external-s3 type
                        Optional("readonly"): bool,  # for external-s3 type
                        Optional("shared"): bool,
                    },
                ],
                Optional("overlapping_secrets"): [str],
                "secrets": {
                    Optional(str): str,
                },
                "env_vars": {
                    Optional(str): str,
                },
            },
        ],
    },
)


def get_paas_env_vars(client: CloudFoundryClient, paas: str) -> dict:
    org, space, app = paas.split("/")

    env_vars = None

    for paas_org in client.v2.organizations:
        if paas_org["entity"]["name"] == org:
            for paas_space in paas_org.spaces():
                if paas_space["entity"]["name"] == space:
                    for paas_app in paas_space.apps():
                        if paas_app["entity"]["name"] == app:
                            env_vars = paas_app["entity"]["environment_json"]

    if not env_vars:
        raise Exception(f"Application {paas} not found")

    return dict(env_vars)


def load_and_validate_config(path):
    with open(path, "r") as fd:
        conf = yaml.safe_load(fd)

    # validate the file
    schema = Schema(config_schema)
    config = schema.validate(conf)

    return config


@click.group()
def bootstrap():
    pass


@bootstrap.command()
def make_config():
    """Generate Copilot boilerplate code."""

    base_path = Path(".")
    config = load_and_validate_config("bootstrap.yml")

    templates = setup_templates()

    click.echo(">>> Generating Copilot configuration files\n")

    # create copilot directory
    mkdir(base_path, "copilot")

    # create copilot/.workspace file
    contents = f"application: {config['app']}"
    click.echo(mkfile(base_path, "copilot/.workspace", contents))

    # create copilot/environments directory
    mkdir(base_path, "copilot/environments")

    # create each environment directory and manifest.yml
    for name, env in config["environments"].items():
        mkdir(base_path, f"copilot/environments/{name}")
        contents = templates.get_template("env/manifest.yml").render(
            {
                "name": name,
                "certificate_arn": env["certificate_arns"][0] if "certificate_arns" in env else "",
            },
        )
        click.echo(mkfile(base_path, f"copilot/environments/{name}/manifest.yml", contents))

    # create each service directory and manifest.yml
    for service in config["services"]:
        service["ipfilter"] = any(
            env.get("ipfilter", False) for _, env in service["environments"].items()
        )
        name = service["name"]
        mkdir(base_path, f"copilot/{name}/addons/")

        if "secrets_from" in service:
            # Copy secrets from the app referred to in the "secrets_from" key
            related_service = [
                s for s in config["services"] if s["name"] == service["secrets_from"]
            ][0]

            service["secrets"].update(related_service["secrets"])

        contents = templates.get_template(f"svc/manifest-{service['type']}.yml").render(service)

        click.echo(mkfile(base_path, f"copilot/{name}/manifest.yml", contents))

    # link to GitHub docs
    click.echo(
        "\nGitHub documentation: "
        "https://github.com/uktrade/platform-documentation/blob/main/gov-pass-to-copilot-migration",
    )


@bootstrap.command()
@click.option("--project-profile", required=True, help="aws account profile name")
@click.option("--env", help="Migrate secrets from a specific environment")
@click.option("--svc", help="Migrate secrets from a specific service")
@click.option(
    "--overwrite",
    is_flag=True,
    show_default=True,
    default=False,
    help="Overwrite existing secrets?",
)
@click.option("--dry-run", is_flag=True, show_default=True, default=False, help="dry run")
def migrate_secrets(project_profile, env, svc, overwrite, dry_run):
    """
    Migrate secrets from your gov paas application to AWS/copilot.

    You need to be authenticated via cf cli and the AWS cli to use this commmand.

    If you're using AWS profiles, use the AWS_PROFILE env var to indicate the which profile to use, e.g.:

    AWS_PROFILE=myaccount copilot-bootstrap.py ...
    """

    check_aws_conn(project_profile)
    # TODO: optional SSM or secret manager

    cf_client = CloudFoundryClient.build_from_cf_config()
    config_file = "bootstrap.yml"
    config = load_and_validate_config(config_file)

    if env and env not in config["environments"].keys():
        raise click.ClickException(f"{env} is not an environment in {config_file}")

    if svc and svc not in [service["name"] for service in config["services"]]:
        raise click.ClickException(f"{svc} is not a service in {config_file}")

    existing_ssm_data = defaultdict(list)
    for env_name, _ in config["environments"].items():
        if env and env_name != env:
            continue

        existing_ssm_data[env_name] = get_ssm_secret_names(config["app"], env_name)

    # get the secrets from the paas
    for service in config["services"]:
        service_name = service["name"]
        secrets = service["secrets"]

        if svc and service["name"] != svc:
            continue

        if "secrets_from" in service:
            click.echo(f"{service_name} shares secrets with {service['secrets_from']}; skipping")
            continue

        for env_name, environment in service["environments"].items():
            if env and env_name != env:
                continue

            click.echo("-----------------")
            click.echo(
                f">>> migrating secrets for service: {service_name}; environment: {env_name}"
            )

            click.echo(f"getting env vars for from {environment['paas']}")
            env_vars = get_paas_env_vars(cf_client, environment["paas"])

            click.echo("Transfering secrets ...")
            for app_secret_key, ssm_secret_key in secrets.items():
                ssm_path = SSM_PATH.format(app=config["app"], env=env_name, name=ssm_secret_key)

                if app_secret_key not in env_vars:
                    # NOT FOUND
                    param_value = "NOT FOUND"
                    click.echo(
                        f"Key not found in paas app: {app_secret_key}; setting to 'NOT FOUND'"
                    )
                elif not env_vars[app_secret_key]:
                    # FOUND BUT EMPTY STRING
                    param_value = "EMPTY"
                    click.echo(
                        f"Empty env var in paas app: {app_secret_key}; SSM requires a non-empty string; setting to 'EMPTY'",
                    )
                else:
                    param_value = env_vars[app_secret_key]

                param_exists = ssm_path in existing_ssm_data[env_name]

                if overwrite or not param_exists:
                    if not dry_run:
                        set_ssm_param(
                            config["app"], env_name, ssm_path, param_value, overwrite, param_exists
                        )

                    if not param_exists:
                        existing_ssm_data[env_name].append(ssm_path)

                if not dry_run:
                    text = (
                        "Created"
                        if not param_exists
                        else "Overwritten"
                        if overwrite
                        else "NOT overwritten"
                    )
                    click.echo(f"{text} {ssm_path}")
                else:
                    click.echo(f"{ssm_path} not created because `--dry-run` flag was included.")


@bootstrap.command()
@click.argument("source_environment")
@click.argument("target_environment")
@click.option("--project-profile", required=True, help="AWS account profile name")
def copy_secrets(project_profile, source_environment, target_environment):
    """Copy secrets from one environment to a new environment."""
    check_aws_conn(project_profile)

    if not Path(f"copilot/environments/{target_environment}").exists():
        click.echo(f"""Target environment manifest for "{target_environment}" does not exist.""")
        exit(1)

    config_file = "bootstrap.yml"
    config = load_and_validate_config(config_file)
    secrets = get_ssm_secrets(config["app"], source_environment)

    for secret in secrets:
        secret_name = secret[0].replace(f"/{source_environment}/", f"/{target_environment}/")

        click.echo(secret_name)

        try:
            set_ssm_param(
                config["app"],
                target_environment,
                secret_name,
                secret[1],
                False,
                False,
                f"Copied from {source_environment} environment.",
            )
        except ClientError as e:
            if e.response["Error"]["Code"] == "ParameterAlreadyExists":
                click.secho(
                    f"""The "{secret_name.split("/")[-1]}" parameter already exists for the "{target_environment}" environment.""",
                    fg="yellow",
                )
            else:
                raise e


if __name__ == "__main__":
    bootstrap()
